// Time-stamp: "23.06.11 06:18 LCD44780.h klaus?wachtler.de"
// $Id: LCD44780.h 620 2011-07-01 05:04:42Z klaus $
//
// Ansteuerung eines HD44780 kompatiblen LCD im 4-Bit-Interfacemodus
//
// [1] http://www.mikrocontroller.net/articles/AVR-GCC-Tutorial#LCD-Ansteuerung
//
// [2] http://www.mikrocontroller.net/topic/88543#751982
//
// [3] "DOTMATRIXDISPLAYS 4x40" von
//     Electronic Assembly GmbH
//     Lochhamer Schlag 17
//     D-82166 Gräfelfing
//     http://www.lcd-module.de
//
// [4] "BLUE LINE"
//     "BLAU NEGATIV MIT WEISSER LED"
//     "DOTMATRIX LCD-Serie 1x16..4x40"
//     Manual beigelegt zu blauen LCDs von Electronic Assembly GmbH
//     Lochhamer Schlag 17
//     D-82166 Gräfelfing
//     http://www.lcd-module.de
// [5] TC1602A-08
//     http://www.abcd8.com/MANNUL/YC162A.pdf
//
// Gegenueber dem Original von [1] sind folgende wesentlichen
// Aenderungen vorgenommen worden:
// - Es kann zwanglos mehr als ein LCD gleichzeitig benutzt werden. Dabei
//   koennen alle ausser der/den Enable-Leitung(en) gemeinsam benutzt
//   werden.
// - Die Signale koennen frei ueber alle Ports verteilt werden.
// - Bei der Initialisierung muss der LCD-Typ angegeben werden.
//   Als Lohn der Muehe sollte die Positionierung auf allen Typen
//   funktionieren, wobei aber nicht alle getestet sind.
// - Der aufrufende Quelltext muss in C++ geschrieben sein, weil sonst
//   keine Referenzen und Defaultparameter moeglich waeren; zudem
//   werden die Methoden dadurch inline.
// - Sowohl Spalten als auch Zeilen werden ab 0 gezaehlt
// - Getestet mit 4x40 mit 2 Controllern
// - printf()-aehnliche Funktionen
// - selbstdefinierte Zeichen moeglich
// - ..._P-Funktionen zum Lesen aus Programmspeicher
//
//
// Bei der Auswahl des Typs (LCD_Type_...) muß man etwas aufpassen,
// wenn man
// - ein einzeiliges mit 16 Zeichen hat, da gibt es zwei Arten:
//   a) MUX 1:8, oder
//   b) MUX 1:16 (8+8); das verhält sich wie ein 2x8, bei dem die
//      beiden Zeilen hintereinander gehängt sind.
//   Dementsprechend ist LCD_Type_1x16_MUX8 oder LCD_Type_1x16_MUX16
//   auszuwählen.
// - ein 4x20 hat. Dann kann es sich einen der Controller HD44780 oder
//   K0073 oder HD66712 handeln.
//   Bei denen unterscheidet sich bei 4x20 die Adressierung (TODO,
//   siehe unten)
//
//
// Beispielverwendung: siehe AVR_LCD44780.cpp.
//
//
// Aenderungen:
//
// 23.06.2009 kw    - von aussen nicht aufzurufende Funktionen static
//                    gemacht:
//                    lcd_data(), lcd_command(), lcd_enable()
//                  - gemeinsamen Code aus lcd_command() und lcd_data()
//                    in neue Funktion lcd_byte() ausgelagert
//                  - Shiften und Maskieren in lcd_byte() eliminiert
//                  - in lcd_set_cursor() Adressberechnung korrigiert
//                    fuer blaues LCD "Blue Line" von Electronic Assembly
//                    EA W204B-NLW bzw. Reichelt LCD204BBL
//
// 02.08.2009 kw    - auf C++ umgestellt
//                  - Ports und Pinnummern beim Initialisieren
//                    anzugeben (anstatt fest verdrahtet)
//                  - mehrere LCDs nutzbar (nur Enable muss auf
//                    unterschiedlichen Pins liegen; fuer ein Display
//                    braucht man also 4 Datenpins, RS und ein oder
//                    zwei Enable; fuer jedes weitere LCD
//                    ein oder zwei weitere Enable).
//                  - Zeilen zaehlen jetzt ebenso wie Spalten ab 0
//
// 26.08.2009 kw    - Parameter fuer LCD-Typ eingefuehrt
//                  - 4x40 eingebaut und getestet
//
// 28.08.2009 kw    - selbst definierte Zeichen mit defChar()
//
// 29.08.2009 kw    - es werden intern Zeilen- und Spaltennummern
//                    mitgezählt und am Zeilenende mit Schreiben
//                    abgebrochen.
//
// 21.05.2010 kw    LCD_Type_4x20_KS0073 und LCD_Type_4x20_HD66712 ergänzt
//
//
// Getestet mit
// - LCD:
//   - Electronic Assembly EA W204B-NLW 4x20 (LCD_Type_4x20)
//   - 2x16 aus Pollin Addon Board zu Atmel Evaluation Board (LCD_Type_2x16)
//   - 2x20
//   - 4x40 (Displaytech 404B, LCD_Type_4x40)
// - Compiler: avr-g++ 4.3.2 (Debian Linux)
//

#ifndef __AnyWare_LCD44780_h__
#define __AnyWare_LCD44780_h__

#include <inttypes.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <alloca.h>
#include <util/delay.h>


#include <avr/eeprom.h>

namespace AnyWare
{

  class LCD44780
  {
  public:


    enum e_LCD_State
      {
        LCD_State_OK,
        LCD_State_TypeInvalid,
        LCD_State_PinInvalid,
        LCD_State_ColumnNumberInvalid,
        LCD_State_LineNumberInvalid,
        LCD_State_WidthTooSmall,
        LCD_State_Unknown
      };


    enum e_LCD_Type
      {
        LCD_Type_1x8,
        LCD_Type_1x16_MUX8,
        LCD_Type_1x16_MUX16,
        LCD_Type_1x20,
        LCD_Type_1x40,
        LCD_Type_2x8,
        LCD_Type_2x12,
        LCD_Type_2x16,
        LCD_Type_2x20,
        LCD_Type_2x24,
        LCD_Type_2x40,
        LCD_Type_4x16,
        // HD44780:
        LCD_Type_4x20,
        LCD_Type_4x20_KS0073,
        // 4x20 mit Controller HD66712, z.B. EA P204-3N/-3NLED
        // evtl. Batron BT42003 "Micro-Line"
        LCD_Type_4x20_HD66712,
        LCD_Type_4x40,
        LCD_Type_invalid
      };


    LCD44780( e_LCD_Type LCD_Type,
              volatile uint8_t &port_d4, uint8_t pin_d4,
              volatile uint8_t &port_d5, uint8_t pin_d5,
              volatile uint8_t &port_d6, uint8_t pin_d6,
              volatile uint8_t &port_d7, uint8_t pin_d7,
              volatile uint8_t &port_rs, uint8_t pin_rs,
              volatile uint8_t &port_e1, uint8_t pin_e1,
              volatile uint8_t &port_e2 = *(uint8_t*)0, uint8_t pin_e2 = (uint8_t)-1
              )
      : LCD_State( LCD_State_Unknown ),
        LCD_Type( LCD_Type ),
        iZeile( 0 ),
        iSpalte( 0 ),
        port_e1( port_e1 ),
        pin_e1( pin_e1 ),
        port_e2( port_e2 ),
        pin_e2( pin_e2 ),
        port_rs( port_rs ),
        pin_rs( pin_rs ),
        port_d4( port_d4 ),
        pin_d4( pin_d4 ),
        port_d5( port_d5 ),
        pin_d5( pin_d5 ),
        port_d6( port_d6 ),
        pin_d6( pin_d6 ),
        port_d7( port_d7 ),
        pin_d7( pin_d7 )
    {
      switch( LCD_Type )
      {
        case LCD_Type_1x8:
          nZeilen  = 1;
          nSpalten = 8;
          map_Linestart_DDRAM[0] = 0x80 + 0x00;
          map_Linestart_DDRAM[1] = 0x80 + 0x00;
          map_Linestart_DDRAM[2] = 0x80 + 0x00;
          map_Linestart_DDRAM[3] = 0x80 + 0x00;
          break;

        case LCD_Type_1x16_MUX8:
          // Anm.: die Zeichen liegen bei 0x00 bis 0x0F
          nZeilen  = 1;
          nSpalten = 16;
          map_Linestart_DDRAM[0] = 0x80 + 0x00;
          map_Linestart_DDRAM[1] = 0x80 + 0x00;
          map_Linestart_DDRAM[2] = 0x80 + 0x00;
          map_Linestart_DDRAM[3] = 0x80 + 0x00;
          break;

        case LCD_Type_1x16_MUX16:
          // Anm. die ersten 8 Zeichen liegen bei 0x00 bis 0x07, die
          // rechten 8 bei 0x40 bis 0x47.
          // Deshalb Sonderbehandlung noetig; ausserdem kein Schreiben von
          // Strings ueber diese Grenze hinweg moeglich!
          nZeilen  = 1;
          nSpalten = 16;
          map_Linestart_DDRAM[0] = 0x80 + 0x00;
          map_Linestart_DDRAM[1] = 0x80 + 0x00;
          map_Linestart_DDRAM[2] = 0x80 + 0x00;
          map_Linestart_DDRAM[3] = 0x80 + 0x00;
          break;

        case LCD_Type_1x20:
          nZeilen  = 1;
          nSpalten = 20;
          map_Linestart_DDRAM[0] = 0x80 + 0x00;
          map_Linestart_DDRAM[1] = 0x80 + 0x00;
          map_Linestart_DDRAM[2] = 0x80 + 0x00;
          map_Linestart_DDRAM[3] = 0x80 + 0x00;
          break;

        case LCD_Type_1x40:
          nZeilen  = 1;
          nSpalten = 40;
          map_Linestart_DDRAM[0] = 0x80 + 0x00;
          map_Linestart_DDRAM[1] = 0x80 + 0x00;
          map_Linestart_DDRAM[2] = 0x80 + 0x00;
          map_Linestart_DDRAM[3] = 0x80 + 0x00;
          break;

        case LCD_Type_2x8:
          nZeilen  = 2;
          nSpalten = 8;
          map_Linestart_DDRAM[0] = 0x80 + 0x00;
          map_Linestart_DDRAM[1] = 0x80 + 0x00;
          map_Linestart_DDRAM[2] = 0x80 + 0x00;
          map_Linestart_DDRAM[3] = 0x80 + 0x00;
          break;

        case LCD_Type_2x12:
          nZeilen  = 2;
          nSpalten = 12;
          map_Linestart_DDRAM[0] = 0x80 + 0x00;
          map_Linestart_DDRAM[1] = 0x80 + 0x40;
          map_Linestart_DDRAM[2] = 0x80 + 0x00;
          map_Linestart_DDRAM[3] = 0x80 + 0x00;
          break;

        case LCD_Type_2x16:
          nZeilen  = 2;
          nSpalten = 16;
          map_Linestart_DDRAM[0] = 0x80 + 0x00;
          map_Linestart_DDRAM[1] = 0x80 + 0x40;
          map_Linestart_DDRAM[2] = 0x80 + 0x00;
          map_Linestart_DDRAM[3] = 0x80 + 0x00;
          break;

        case LCD_Type_2x20:
          nZeilen  = 2;
          nSpalten = 20;
          map_Linestart_DDRAM[0] = 0x80 + 0x00;
          map_Linestart_DDRAM[1] = 0x80 + 0x40;
          map_Linestart_DDRAM[2] = 0x80 + 0x00;
          map_Linestart_DDRAM[3] = 0x80 + 0x00;
          break;

        case LCD_Type_2x24:
          nZeilen  = 2;
          nSpalten = 24;
          map_Linestart_DDRAM[0] = 0x80 + 0x00;
          map_Linestart_DDRAM[1] = 0x80 + 0x40;
          map_Linestart_DDRAM[2] = 0x80 + 0x00;
          map_Linestart_DDRAM[3] = 0x80 + 0x00;
          break;

        case LCD_Type_2x40:
          nZeilen  = 2;
          nSpalten = 40;
          map_Linestart_DDRAM[0] = 0x80 + 0x00;
          map_Linestart_DDRAM[1] = 0x80 + 0x40;
          map_Linestart_DDRAM[2] = 0x80 + 0x00;
          map_Linestart_DDRAM[3] = 0x80 + 0x00;
          break;

        case LCD_Type_4x16:
          nZeilen  = 4;
          nSpalten = 16;
          map_Linestart_DDRAM[0] = 0x80 + 0x00;
          map_Linestart_DDRAM[1] = 0x80 + 0x40;
          map_Linestart_DDRAM[2] = 0x80 + 0x10;
          map_Linestart_DDRAM[3] = 0x80 + 0x50;
          break;

        case LCD_Type_4x20:

          // TODO:
          // EA DIP 204-6 (KS0073): Adressen 0x20 0x40 0x60 0x80
          // beim Initialisieren noch dazu schicken: 0b00001001 (mit RE=1).

          nZeilen  = 4;
          nSpalten = 20;
          map_Linestart_DDRAM[0] = 0x80 + 0x00;
          map_Linestart_DDRAM[1] = 0x80 + 0x40;
          map_Linestart_DDRAM[2] = 0x80 + 0x14;
          map_Linestart_DDRAM[3] = 0x80 + 0x54;
          break;

        case LCD_Type_4x20_KS0073:
          nZeilen  = 4;
          nSpalten = 20;
          map_Linestart_DDRAM[0] = 0x80 + 0x00;
          map_Linestart_DDRAM[1] = 0x80 + 0x20;
          map_Linestart_DDRAM[2] = 0x80 + 0x40;
          map_Linestart_DDRAM[3] = 0x80 + 0x60;
          break;

        case LCD_Type_4x20_HD66712:
          nZeilen  = 4;
          nSpalten = 20;
          map_Linestart_DDRAM[0] = 0x80 + 0x00;
          map_Linestart_DDRAM[1] = 0x80 + 0x20;
          map_Linestart_DDRAM[2] = 0x80 + 0x40;
          map_Linestart_DDRAM[3] = 0x80 + 0x60;
          break;

        case LCD_Type_4x40:
          nZeilen  = 4;
          nSpalten = 40;
          map_Linestart_DDRAM[0] = 0x80 + 0x00;
          map_Linestart_DDRAM[1] = 0x80 + 0x40;
          map_Linestart_DDRAM[2] = 0x80 + 0x00;
          map_Linestart_DDRAM[3] = 0x80 + 0x40;
          break;

        default :
          LCD_State = LCD_State_TypeInvalid;
          return;
          break;

      }

      selectController( 0 );

      // Ports auf Ausgang schalten
      // Anm.: die DDR fuer alle Ports sind jeweils mit einer um 1
      // niedrigeren Adresse eingeblendet als die PORT...; solange Atmel
      // das nicht aendert, kann man ueber (&port_...)[-1] auf das
      // DDR-Register zugreifen.
      (&port_d4)[-1] |= (1<<pin_d4);
      (&port_d5)[-1] |= (1<<pin_d5);
      (&port_d6)[-1] |= (1<<pin_d6);
      (&port_d7)[-1] |= (1<<pin_d7);
      (&port_e1)[-1] |= (1<<pin_e1);
      if( &port_e2!=(uint8_t*)0 )
      {
        (&port_e2)[-1] |= (1<<pin_e2);
      }
      (&port_rs)[-1] |= (1<<pin_rs);

      // muss 3 mal hintereinander gesendet werden zur Initialisierung

      _delay_ms( 15 );

      port_d4 |=  (1<<pin_d4);
      port_d5 |=  (1<<pin_d5);
      port_d6 &= ~(1<<pin_d6);
      port_d7 &= ~(1<<pin_d7);

      port_rs &= ~(1<<pin_rs);      // RS auf 0
      enable();
      if( &port_e2!=(uint8_t*)0 )
      {
        selectController( 1 );
        enable();
        selectController( 0 );
      }

      _delay_ms( 5 );
      enable();
      if( &port_e2!=(uint8_t*)0 )
      {
        selectController( 1 );
        enable();
        selectController( 0 );
      }

      _delay_ms( 1 );
      enable();
      if( &port_e2!=(uint8_t*)0 )
      {
        selectController( 1 );
        enable();
        selectController( 0 );
      }
      _delay_ms( 1 );

      // 4 Bit Modus aktivieren
      port_d4 &=  ~(1<<pin_d4);
      port_d5 |=   (1<<pin_d5);
      port_d6 &=  ~(1<<pin_d6);
      port_d7 &=  ~(1<<pin_d7);

      enable();
      if( &port_e2!=(uint8_t*)0 )
      {
        selectController( 1 );
        enable();
        selectController( 0 );
      }
      _delay_ms( 1 );

      // 4Bit / 2 Zeilen / 5x7
      // TODO: für 1x** ändern?
      // TODO: Wieso sind auf dem EA W204B mehr als 7 Pixelzeilen zu sehen?
      writeCommand( 0x28 );
      if( &port_e2!=(uint8_t*)0 )
      {
        selectController( 1 );
        writeCommand( 0x28 );
        selectController( 0 );
      }

      // Display ein / Cursor aus / kein Blinken
      writeCommand( 0x0C );
      if( &port_e2!=(uint8_t*)0 )
      {
        selectController( 1 );
        writeCommand( 0x0C );
        selectController( 0 );
      }

      // inkrement / kein Scrollen
      writeCommand( 0x06 );
      if( &port_e2!=(uint8_t*)0 )
      {
        selectController( 1 );
        writeCommand( 0x06 );
        selectController( 0 );
      }

      clear();
      LCD_State = LCD_State_OK;
    } // LCD44780()


    // Loescht letzte Fehlernummer:
    void resetState()
    {
      LCD_State = LCD_State_OK;
    }

    // liefert den Status bzw. den letzten Fehler
    e_LCD_State getState() const
    {
      return LCD_State;
    }


    // liefert aktuelle Zeilennummer
    uint8_t getLineNumber() const
    {
      return iZeile;
    }


    // liefert aktuelle Spaltennummer
    uint8_t getColumnNumber() const
    {
      return iSpalte;
    }


    // Loescht das Display
    void clear()
    {
      if( LCD_Type==LCD_Type_4x40 )
      {
        selectController( 0 );
      }
      writeCommand( LCD_CLEAR_DISPLAY );
      if( LCD_Type==LCD_Type_4x40 )
      {
        selectController( 1 );
        writeCommand( LCD_CLEAR_DISPLAY );
        selectController( 0 );
      }
      _delay_ms( 5 );
      iZeile = iSpalte = 0;
    }


    // Sendet den Befehl: Cursor Home
    void home()
    {
      if( LCD_Type==LCD_Type_4x40 )
      {
        selectController( 0 );
      }
      writeCommand( LCD_CURSOR_HOME );
      _delay_ms( 5 );
      iZeile = iSpalte = 0;
    }


    // setzt den Cursor in Zeile y (0..3) Spalte x (0..39)
    void setCursor( uint8_t x, uint8_t y )
    {
      if( x>=nSpalten )
      {
        LCD_State = LCD_State_ColumnNumberInvalid;
        return;
      }

      if( y>=nZeilen )
      {
        LCD_State = LCD_State_LineNumberInvalid;
        return;
      }

      if( LCD_Type==LCD_Type_4x40 )
      {
        // Zeilen 0 und 1 werden vom ersten Controller (0) verwurstet,
        // die Zeilen 2 und 3 vom zweiten (1).
        selectController( y>=2 );
      }

      // Sonderbehandlung fuer LCD_Type_1x16_MUX16, wenn x>=8
      if( LCD_Type==LCD_Type_1x16_MUX16 && x>=8 )
      {
        writeCommand( 0x40 + x - 8 );
      }
      else
      {
        writeCommand( map_Linestart_DDRAM[y] + x );
      }

      iSpalte = x;
      iZeile = y;
    }

    // Schreibt einen nullterminierten String auf das LCD
    void printSZ( const char *p_data )
    {
      while( p_data && *p_data )
      {
        printChar( *p_data++ );
      }
    }


    // Schreibt einen String aus dem Flash auf das LCD
    void printSZ_P( const char *p_data )
    {
      char   data;
      while( p_data && (data = (char)eeprom_read_byte( ((const uint8_t*)p_data) )) )
      {
        p_data++;
        printChar( data );
      }
    }


    // Schreibt einen String an der Stelle (x,y) auf das LCD
    void printSZ( uint8_t x, uint8_t y, const char *p_data )
    {
      setCursor( x, y );
      while( *p_data )
      {
        printChar( *p_data++ );
      }
    }


    // Schreibt einen String aus dem Flash auf das LCD
    void printSZ_P( uint8_t x, uint8_t y, const char *p_data )
    {
      char   data;
      setCursor( x, y );
      while( (data = (char)eeprom_read_byte( ((const uint8_t*)p_data) )) )
      {
        p_data++;
        printChar( data );
      }
    }


    // sendet ein Datenbyte an das LCD
    void printChar( unsigned char data )
    {
      if( iSpalte<nSpalten )
      {
        port_rs |= (1<<pin_rs);        // RS auf 1 setzen
        writeByte( data );
        ++iSpalte;
      }
      else
      {
        LCD_State = LCD_State_ColumnNumberInvalid;
      }
    }


    // setzt Cursor und sendet ein Datenbyte an das LCD
    void printChar( uint8_t x, uint8_t y, unsigned char data )
    {
      setCursor( x, y );
      printChar( data );
    }

    // definiert ein Zeichen im Zeichensatz an der Position pos (0 bis
    // 7) mit den Bitdaten aus arrBitmuster[].
    // arrBitmuster[] enthaelt 8 uint8_t, wobei das erste Byte in
    // seinen unteren 4 Bit die oberste Zeile des 5x8 Pixel grossen
    // Zeichens enthaelt usw. bis arrBitmuster[7], das in seinen
    // unteren 5 Bit die unterste Pixelzeile enthaelt.
    // Bei einem 2x16 aus dem Pollin Addon-Board kann ich nur die
    // ersten 4 Zeichen definieren; bei allen anderen waren es bisher
    // immer 8.
    void defChar( uint8_t pos, uint8_t arrBitmuster[8] )
    {
      uint8_t    *pBitmuster = arrBitmuster;
      // Kommando set CGRAM address gem. [3]: binaer 01<pos:3><Pixelzeile:3>
      uint8_t  cmdSetCGRAMAdress = 0b01000000 | ((pos&0b11) << 3);

      for( uint8_t iPixelzeile=0; iPixelzeile<8; ++iPixelzeile )
      {
        writeCommand( cmdSetCGRAMAdress++ ); printChar( *pBitmuster++ );
        _delay_ms( 5 );
      }

      // ggf. fuer den anderen Controller auch:
      if( LCD_Type==LCD_Type_4x40 )
      {
        changeController();
        pBitmuster = arrBitmuster;
        cmdSetCGRAMAdress = 0b01000000 | ((pos&0b11) << 3);
        for( uint8_t iPixelzeile=0; iPixelzeile<8; ++iPixelzeile )
        {
          writeCommand( cmdSetCGRAMAdress++ ); printChar( *pBitmuster++ );
        }
        changeController();
      }

      // Mit dem Definieren eines Zeichens scheint zumindest der 2x16
      // vom Pollin Addon-Board seine Ausgabeposition zu vergessen.
      // Neu setzen:
      setCursor( iSpalte, iZeile );
    }


    // gibt an der aktuellen Cursorposition aus, sonst wie
    // std::printf().
    // Rueckgabe: Anzahl der ausgegebenen Zeichen (0, wenn es mehr als
    // geworden worden waeren, als in die Zeile passt; in diesem Fall
    // werden '#' ausgegeben)
    uint8_t printf( const char *fmt, ... )
    {
      uint8_t   ret = 0;
      uint8_t   restlicheZeilenlaenge = nSpalten - iSpalte;
      char     *puffer = (char*)alloca( restlicheZeilenlaenge + 1 );
      va_list   args;
      va_start( args, fmt );
      ret = vsnprintf( puffer, restlicheZeilenlaenge+1, fmt, args );
      if( ret>restlicheZeilenlaenge )
      {
        for( uint8_t i=iSpalte; i<nSpalten; ++i )
        {
          printChar( '#' );
        }
        ret = 0;
        LCD_State = LCD_State_ColumnNumberInvalid;
      }
      else
      {
        printSZ( puffer );
      }
      va_end( args );
      return ret;
    }


    // gibt an der aktuellen Cursorposition aus, sonst wie
    // std::printf().
    // fmt ist ein Zeiger in den Programmspeicher.
    // Rueckgabe: Anzahl der ausgegebenen Zeichen (0, wenn es mehr als 40
    // geworden worde waeren, als in die Zeile passt; in diesem Fall
    // werden '#' ausgegeben)
    uint8_t printf_P( const char *fmt, ... )
    {
      // maximale Laenge fuer ein LCD ist 40, eins fuer die Null:
      uint8_t   ret = 0;
      uint8_t   restlicheZeilenlaenge = nSpalten - iSpalte;
      char     *puffer = (char*)alloca( restlicheZeilenlaenge + 1 );
      va_list   args;
      va_start( args, fmt );
      ret = vsnprintf_P( puffer, sizeof(puffer), fmt, args );
      if( ret>restlicheZeilenlaenge )
      {
        for( uint8_t i=iSpalte; i<nSpalten; ++i )
        {
          printChar( '#' );
        }
        ret = 0;
        LCD_State = LCD_State_ColumnNumberInvalid;
      }
      else
      {
        printSZ( puffer );
      }
      va_end( args );
      return ret;
    }


    // gibt an der Cursorposition (x,y) n Zeichen aus, sonst wie std::printf().
    // Wenn die Ausgabe gem. fmt nicht zu n Zeichen fuehrt, wird rechts mit
    // Leerzeichen aufgefuellt.
    // Rueckgabe: Anzahl der ausgegebenen Zeichen ohne die aufgefuellten
    // Leerzeichen (0, wenn es mehr als n geworden waeren oder wenn
    // n>Zeilenlaenge; in diesen Faellen werden '#' ausgegeben)
    uint8_t printf( uint8_t x, uint8_t y, uint8_t n, const char *fmt, ... )
    {
      uint8_t   ret = 0;
      char     *puffer = (char *)alloca( n + 1 );
      va_list   args;
      char      fill = '#';

      setCursor( x, y );

      if( n<=( nSpalten - x ) )
      {
        va_start( args, fmt );
        ret = vsnprintf( puffer, n+1, fmt, args );
        if( ret>n )
        {
          // Fehler, mehr als n Zeichen wuerden ausgegeben werden
          LCD_State = LCD_State_WidthTooSmall;
          ret = 0;
        }
        else
        {
          // Formatierten String ausgeben:
          printSZ( puffer );
          fill = ' ';
        }
        va_end( args );
      }
      else
      {
        n = nSpalten - x;
      }

      // Ggf. mit ' ' oder '#' auffuellen:
      for( uint8_t i=ret; i<n; ++i )
      {
        printChar( fill );
      }
      return ret;
    }


    // gibt an der Cursorposition (x,y) n Zeichen aus, sonst wie std::printf().
    // Wenn die Ausgabe gem. fmt nicht zu n Zeichen fuehrt, wird rechts mit
    // Leerzeichen aufgefuellt.
    // fmt ist ein Zeiger in den Programmspeicher.
    // Rueckgabe: Anzahl der ausgegebenen Zeichen ohne die aufgefuellten
    // Leerzeichen (0, wenn es mehr als n geworden waeren oder wenn
    // n>Zeilenlaenge; in diesen Faellen werden '#' ausgegeben)
    int printf_P( uint8_t x, uint8_t y, uint8_t n, const char *fmt, ... )
    {
      // maximale Laenge fuer ein LCD ist 40, eins fuer die Null:
      int       ret = 0;
      char     *puffer = (char *)alloca( n + 1 );
      va_list   args;
      char      fill = '#';

      setCursor( x, y );

      if( n<=( nSpalten - x ) )
      {
        va_start( args, fmt );
        ret = vsnprintf_P( puffer, n+1, fmt, args );
        if( ret>n )
        {
          // Fehler, mehr als n Zeichen wuerden ausgegeben werden
          LCD_State = LCD_State_WidthTooSmall;
          ret = 0;
        }
        else
        {
          // Formatierten String ausgeben:
          printSZ( puffer );
          fill = ' ';
        }
        va_end( args );
      }
      else
      {
        n = nSpalten - x;;
      }

      // Ggf. mit ' ' oder '#' auffuellen:
      for( uint8_t i=ret; i<n; ++i )
      {
        printChar( fill );
      }
      return ret;
    }


  private:


    // waehlt den ersten (0) oder zweiten (1) Controller aus
    void selectController( uint8_t iCtrlr )
    {
      switch( iCtrlr )
      {
        case 0:
          p_port_e = &port_e1;
          p_pin_e  = &pin_e1;
          break;

        case 1:
          p_port_e = &port_e2;
          p_pin_e  = &pin_e2;
          break;

        default :

          break;
      }
    }


    // wechselt den Controller (1 , wenn vorher 0; 0 wenn vorher 1)
    void changeController()
    {
      if( p_port_e == &port_e2 && p_pin_e== &pin_e2 )
      {
        p_port_e = &port_e1;
        p_pin_e  = &pin_e1;
      }
      else
      {
        p_port_e = &port_e2;
        p_pin_e  = &pin_e2;
      }
    }


    // erzeugt den Enable-Puls fuer den Controller, der zuletzt mit
    // selectController() ausgewaehlt wurde
    void enable() const
    {
      // Bei Problemen ggf. Pause gemaess Datenblatt des LCD Controllers einfuegen

      // http://www.mikrocontroller.net/topic/81974#685882
      *p_port_e |= (1<<*p_pin_e);

      // Bei Problemen ggf. Pause gemaess Datenblatt des LCD Controllers verlaengern
      _delay_us( 1 );                   // kurze Pause

      // http://www.mikrocontroller.net/topic/80900
      *p_port_e &= ~(1<<*p_pin_e);
    }


    // sendet ein Daten- oder Kommandobyte (RS am LCD muss vorher
    // entsprechend gesetzt sein)
    void writeByte( unsigned char data ) const
    {
      // oberes Nibble auf den LCD-Pins 4..7 ausgeben:

      if( data & 0x10 )
      {
        port_d4 |= (1<<pin_d4);
      }
      else
      {
        port_d4 &= ~(1<<pin_d4);
      }
      if( data & 0x20 )
      {
        port_d5 |= (1<<pin_d5);
      }
      else
      {
        port_d5 &= ~(1<<pin_d5);
      }
      if( data & 0x40 )
      {
        port_d6 |= (1<<pin_d6);
      }
      else
      {
        port_d6 &= ~(1<<pin_d6);
      }
      if( data & 0x80 )
      {
        port_d7 |= (1<<pin_d7);
      }
      else
      {
        port_d7 &= ~(1<<pin_d7);
      }
      enable();

      // unteres Nibble auf den LCD-Pins 4..7 ausgeben:

      if( data & 0x01 )
      {
        port_d4 |= (1<<pin_d4);// setzen
      }
      else
      {
        port_d4 &= ~(1<<pin_d4);
      }
      if( data & 0x02 )
      {
        port_d5 |= (1<<pin_d5);
      }
      else
      {
        port_d5 &= ~(1<<pin_d5);
      }
      if( data & 0x04 )
      {
        port_d6 |= (1<<pin_d6);
      }
      else
      {
        port_d6 &= ~(1<<pin_d6);
      }
      if( data & 0x08 )
      {
        port_d7 |= (1<<pin_d7);
      }
      else
      {
        port_d7 &= ~(1<<pin_d7);
      }
      enable();

      // Dem LCD Zeit geben, das Kommando auszufuehren
      _delay_us(42);
    }


    // sendet einen Befehl an das LCD
    void writeCommand( unsigned char command ) const
    {
      port_rs &= ~(1<<pin_rs);        // RS auf 0 setzen
      writeByte( command );
    }


    enum
      {
        LCD_CLEAR_DISPLAY = 0x01,
        LCD_CURSOR_HOME   = 0x02,
      };


    // aktueller Zustand (LCD_State_OK oder ein Fehler):
    e_LCD_State       LCD_State;


    // Typ (4x20 etc.)
    e_LCD_Type        LCD_Type;


    // Anzahl Zeilen, Spalten:
    uint8_t           nZeilen;
    uint8_t           nSpalten;

    // aktuelle Zeile, Spalte:
    uint8_t           iZeile;
    uint8_t           iSpalte;


    // Ansteuerung (die port_... sind Referenzen auf PORTx; die
    // pin_... sind Pinnummern 0...7):
    volatile uint8_t &port_e1;
    uint8_t           pin_e1;
    volatile uint8_t &port_e2;
    uint8_t           pin_e2;
    volatile uint8_t &port_rs;
    uint8_t           pin_rs;
    volatile uint8_t &port_d4;
    uint8_t           pin_d4;
    volatile uint8_t &port_d5;
    uint8_t           pin_d5;
    volatile uint8_t &port_d6;
    uint8_t           pin_d6;
    volatile uint8_t &port_d7;
    uint8_t           pin_d7;


    // jeweils aktuelle Zeiger auf port_e1 und pin_e1 oder port_e2 und pin_e2:
    volatile uint8_t *p_port_e;
    uint8_t          *p_pin_e;


    // Tabelle, welche Zeile an welcher Stelle im DD-RAM beginnt (wird
    // je nach LCD_Type initalisiert):
    uint8_t       map_Linestart_DDRAM[4];

  }; // class LCD44780
}; // namespace AnyWare

#endif /* ifndef __AnyWare_LCD44780_h__ */

// Local Variables:
// mode:C++
// End:
